/******************************************************************************
 *
 * powerctl_teensy.ino
 *
 * This is the program that runs the telescope power control box. It switches
 * power on and/or off for 6 power ports. It also controls 2 dew heaters,
 * either in automatic mode, based on the dew point, or in manual mode with
 * direct entry of the PWM value (0 - 255). Current monitoring, too.
 *
 *****************************************************************************/

#include <Wire.h>
//#include "Adafruit_Si7021.h"
#include "Adafruit_MCP9808.h"
#include "SparkFunHTU21D.h"
#include <EEPROM.h>
#include "TeensyTimerTool.h"
using namespace TeensyTimerTool;

#define LED 13
#define LED_ON 0
#define LED_OFF 1

// Program options 2.
#define OPT_DEW_ENABLE_1 0
#define OPT_DEW_ENABLE_2 1
#define OPT_AUTO_DEW_1 2
#define OPT_AUTO_DEW_2 3

// Command message offsets
#define CMD_OFS 0
#define CMD_TYPE 3
#define CMD_PORT 4
#define CMD_VALUE 6

// Command value types.
#define DT_BOOL 1
#define DT_BYTE 2
#define DT_FLOAT 3

// command and response buffer sizes are the same.
#define BUFFER_SIZE 96

// Timer for PWM the hard way. 0.064 seconds per cycle
// is 256 x 0.000250 seconds. 15.625Hz
#define TIMER_TIME 250
#define TIMER t1
PeriodicTimer t1;

// DMP! command timer
#define DMP_TIME_DELAY 5000

// EEPROM stuff 64 32-byte pages = 2kB
#define EEPROM_SIZE 2048
#define QUEUE_ENTRIES 64
#define EPR_PAGE_SIZE 32
#define EPR_TIME_DELAY 15000
#define eeprom EEPROM

// CUR? Amperage calc period
#define AMP_TIME_DELAY 10000

// Amperage calibration factors
#define CALIBRATION_MULT 1.114
#define CALIBRATION_ADD -0.23

//Adafruit_Si7021 sensor = Adafruit_Si7021();
HTU21D sensor;
Adafruit_MCP9808 lens_temp1 = Adafruit_MCP9808();
Adafruit_MCP9808 lens_temp2 = Adafruit_MCP9808();

uint8_t loop_count;
uint8_t loop_256;
uint8_t pwm_loop;
unsigned long timer_tick;
unsigned long eeprom_tick;
unsigned long amperage_tick;
unsigned long clock_tick;
byte rt_hours = 0;
byte rt_minutes = 0;
byte rt_seconds = 0;

byte state;
bool epr_data_dirty = false;

struct Epr_Data {
    uint32_t seq_no;
    int16_t sig;
    byte options;
    byte options2;
    byte options3;
    byte heater_offset[2];
    byte pwm_min_values[2];
    byte heater_gain[2];
    byte pwm_values[2];
    byte power_delays[6];
    byte spare[6];
};

Epr_Data SETUP_DATA{
        .seq_no = 0,
        .sig = 0x0dca,
        .options = 0,
        .options2 = 0,
        .options3 = 0,
        .heater_offset = {4, 4},
        .pwm_min_values = {0, 0},
        .heater_gain = {50, 50},
        .pwm_values = {0, 0},
        .power_delays = {0, 255, 0, 0, 0, 0},
};

Epr_Data EPR_DATA {
        .seq_no = 0,
        .sig = 0,
        .options = 0,
        .options2 = 0,
        .options3 = 0,
        .heater_offset = {0, 0},
        .pwm_min_values = {0, 0},
        .heater_gain = {0, 0},
        .pwm_values = {0, 0},
        .power_delays = {0, 0, 0, 0, 0, 0},
};

uint32_t last_sequence_no = 0;
uint16_t queue_tail = 0;

float humidity;
float temperature;
float dew_point;
float lens_temperature[] = {0.0, 0.0};
int power_pins[] = {0, 1, 2, 3, 4, 5};
int heater_pins[] = {6, 7};
int current_pwr_pins[] = {A0, A1, A4, A5, A6, A7};
int current_htr_pins[] = {A8, A9};
float current_pwr[] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
float current_htr[] = {0.0, 0.0};
float ah_pwr[] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
float ah_htr[] = {0.0, 0.0};
float voltage = 0.0;
float calibration_mult_pwr[] = {1.041531, 1.065915, 1.078227, 1.065453, 1.043193, 1.048718};
float calibration_add_pwr[] = {-0.100586, -0.000732, -0.020508, -0.019287, -0.043701, 0.011475};
float calibration_mult_htr[] = {1.037915, 1.049714};
float calibration_add_htr[] = {-0.104248, -0.073975};
int voltage_pin = A10;
bool dump_enabled = false;
bool temp_enabled = true;
char port;
bool value_bool;
uint8_t value_byte;
float value_float;

char cmd_buffer[BUFFER_SIZE];
char rsp_buffer[BUFFER_SIZE];
char good_response[] = {"OK\n"};
char version_string[] = {"OK:1.1\n"};

/******************************************************************************
 * 
 * load_eeprom  Find and load the latest copy of settings in EEPROM.
 * 
 *****************************************************************************/

void load_eeprom()
{
    int i;
    Epr_Data ed;
    last_sequence_no = 0;
    queue_tail = 0;

    memcpy(&EPR_DATA, &SETUP_DATA, sizeof(EPR_DATA));

    for (i = 0; i < QUEUE_ENTRIES; i++)
    {
        eeprom.get((i * EPR_PAGE_SIZE), ed);
        if ((ed.seq_no > last_sequence_no) && (ed.seq_no != 0xFFFFFFFF))
        {
            queue_tail = i;
            last_sequence_no = ed.seq_no;
        }
    }
    eeprom.get((queue_tail * EPR_PAGE_SIZE), EPR_DATA);
}

/******************************************************************************
 * 
 * store_eeprom Store the latest copy in EEPROM and increment counter.
 *
 * Writes to a different page each time it writes, incrementing a 32-bit
 * counter.
 *
 *****************************************************************************/

void store_eeprom() {
    queue_tail++;
    if (queue_tail >= QUEUE_ENTRIES)
        queue_tail = 0;
    last_sequence_no++;
    EPR_DATA.seq_no = last_sequence_no;
    eeprom.put(queue_tail * EPR_PAGE_SIZE, EPR_DATA);
}

/******************************************************************************
 *
 * setPortPwr   Turns the power on for this port.
 *
 * @param p     The port number (0-5).
 * @param flag  The port should be powered on - True or false.
 * @return      An error message or response.
 *
 *****************************************************************************/

char * setPortPwr(byte p, bool flag) {
    if (flag) {
        state = bitSet(state, p);
    } else {
        state = bitClear(state, p);
    }
    pinMode(power_pins[p], OUTPUT);
    digitalWrite(power_pins[p], flag ? HIGH : LOW);
    return good_response;
}

/******************************************************************************
 *
 * getPortPwr   Returns the status of the port pin.
 *
 * @param p     The port to query.
 * @return      An error message or response.
 *
 *****************************************************************************/

char * getPortPwr(byte p) {
    strcpy(rsp_buffer, "OK:");
    strcat(rsp_buffer, bitRead(state, p) ? "ON\n" : "OFF\n");
    return rsp_buffer;
}

/******************************************************************************
 *
 * setPortAuto  Sets the port to automatically come on at power up.
 *
 * @param p     The port to set.
 * @param flag  The value to set - True or false.
 * @return      An error message or response.
 *
 *****************************************************************************/

char * setPortAuto(byte p, bool flag) {
    if (flag) {
        EPR_DATA.options = bitSet(EPR_DATA.options, p);
    } else {
        EPR_DATA.options = bitClear(EPR_DATA.options, p);
    }
    epr_data_dirty = true;
    return good_response;
}

/******************************************************************************
 *
 * getPortAuto  Gets the auto start flag for the specified port.
 *
 * @param p     The port to query.
 * @return      An error message or response.
 *
 *****************************************************************************/

char * getPortAuto(byte p) {
    strcpy(rsp_buffer, "OK:");
    strcat(rsp_buffer, bitRead(EPR_DATA.options, p) ? "ON\n" : "OFF\n");
    return rsp_buffer;
}

/******************************************************************************
 *
 * setHeaterPwm     Manually set the heater PWM value, unless auto is set.
 *
 * @param h         The heater number to set.
 * @param pwm_value The pwm value (0 - 255)
 * @return          An error message or response.
 *
 *****************************************************************************/

char * setHeaterPwm(byte h, byte pwm_value) {
    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    if (!bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_1 + h)) {
        strcpy(rsp_buffer, "ER:disabled\n");
        return rsp_buffer;
    } else if (bitRead(EPR_DATA.options2, OPT_AUTO_DEW_1 + h)) {
        strcpy(rsp_buffer, "ER:auto mode\n");
        return rsp_buffer;
    } else {
        EPR_DATA.pwm_values[h] = pwm_value;
        return good_response;
    }
}

/******************************************************************************
 *
 * getHeaterPwm     Returns the heater pwm value (0 - 255)
 *
 * @param h         The heater number.
 * @return          An error message or response.
 *
 *****************************************************************************/

char * getHeaterPwm(byte h) {
    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%03u\n", EPR_DATA.pwm_values[h]);
    return rsp_buffer;
}

/******************************************************************************
 *
 * setHeaterAuto    Sets the mode to manual or automatic for the dew heaters.
 *
 * @param flag      The value to set (true - false)
 * @return          An error message or response.
 *
 *****************************************************************************/

char * setHeaterAuto(byte h, bool flag) {
    byte tmp;

    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    tmp = EPR_DATA.options2;
    if (flag) {
        if (reconcile_htr_auto(h)) {
            EPR_DATA.options2 = bitSet(EPR_DATA.options2, OPT_AUTO_DEW_1 + h);
            strcpy(rsp_buffer, "OK\n");
        } else {
            EPR_DATA.options2 = bitClear(EPR_DATA.options2, OPT_AUTO_DEW_1 + h);
            strcpy(rsp_buffer, "ER:PROBE\n");
        }
    } else {
        EPR_DATA.options2 = bitClear(EPR_DATA.options2, OPT_AUTO_DEW_1 + h);
        strcpy(rsp_buffer, "OK\n");
    }
    epr_data_dirty = true;
    return rsp_buffer;
}

/******************************************************************************
 *
 * getHeaterAuto    Get the heater auto control flag.
 *
 * @return          An error message or response.
 *
 *****************************************************************************/

char * getHeaterAuto(byte h) {
    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    strcpy(rsp_buffer, "OK:");
    strcat(rsp_buffer, bitRead(EPR_DATA.options2, OPT_AUTO_DEW_1 + h) ? "ON\n" : "OFF\n");
    return rsp_buffer;
}

/******************************************************************************
 *
 * setHeaterOffset      Set and remember the heater offset for this heater.
 *
 * @param h     Heater number (0 - 1)
 * @param o     Offset temp in degrees C.
 * @return      An error message or response.
 *
 *****************************************************************************/

char * setHeaterOffset(byte h, byte o) {
    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    if (o != EPR_DATA.heater_offset[h]) {
        EPR_DATA.heater_offset[h] = o;
        epr_data_dirty = true;
    }
    return good_response;
}

/******************************************************************************
 *
 * getHeaterOffset      Get the offset in degrees C for this heater.
 *
 * @param h     Heater number (0 - 1)
 * @return      An error message or response.
 *
 *****************************************************************************/

char * getHeaterOffset(byte h) {
    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK: %03d\n", EPR_DATA.heater_offset[h]);
    return rsp_buffer;
}

/******************************************************************************
 *
 * GetCurrentTemp   Returns the temperature.
 *
 * @return          An error message or response.
 *
 *****************************************************************************/

char * GetCurrentTemp() {
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%-5.2f\n", temperature);
    return rsp_buffer;
}

/******************************************************************************
 *
 * GetCurrentDewPoint   Returns the dew point temperature.
 *
 * @return              An error message or response.
 *
 *****************************************************************************/

char * GetCurrentDewPoint() {
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%-5.2f\n", dew_point);
    return rsp_buffer;
}

/******************************************************************************
 *
 * GetCurrentHumidity   Returns the current relative humidity percent.
 *
 * @return              An error message or response.
 *
 *****************************************************************************/

char * GetCurrentHumidity() {
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%-5.2f\n", humidity);
    return rsp_buffer;
}

/******************************************************************************
 *
 * getLensTemp      Gets the remote temperature
 *
 *****************************************************************************/

char * GetLensTemp(byte h) {
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%-5.2f\n", lens_temperature[h]);
    return rsp_buffer;
}

/******************************************************************************
 *
 * setHeaterMinPwm      Set the minimum PWM value for an on heater.
 *
 * @param h     Heater number (0 - 1)
 * @param val   Minimum PWM value (0 - 255)
 * @return      good response or error message.
 *
 *****************************************************************************/

char * setHeaterMinPwm(byte h, byte val) {
    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    if (EPR_DATA.pwm_min_values[h] != val) {
        EPR_DATA.pwm_min_values[h] = val;
        epr_data_dirty = true;
    }
    return good_response;
}

/******************************************************************************
 *
 * getLensTemp      Gets the remote temperature
 *
 *****************************************************************************/

char * getHeaterMinPwm(byte h) {
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%03d\n", EPR_DATA.pwm_min_values[h]);
    return rsp_buffer;
}

/******************************************************************************
 *
 * setHeaterGain        Set the heater proportional gain  (1 - 99)
 *
 * @param h     Heater number (0 - 1)
 * @param gain  Gain (1 - 99)
 * @return      good response or error message.
 *
 *****************************************************************************/

char * setHeaterGain(byte h, byte gain) {
    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    if (gain < 1 || gain > 99) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Gain range %d\n", h);
        return rsp_buffer;
    }
    if (EPR_DATA.heater_gain[h] != gain) {
        EPR_DATA.heater_gain[h] = gain;
        epr_data_dirty = true;
    }
    return good_response;
}

/******************************************************************************
 *
 * getHeaterGain      Gets the heater proportional gain (1 - 255).
 *
 *****************************************************************************/

char * getHeaterGain(byte h) {
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%03d\n", EPR_DATA.heater_gain[h]);
    return rsp_buffer;
}

/******************************************************************************
 *
 * setPowerDelay        Set the delay before powering up this pin. Delay does
 *                      not apply to low outputs - only high outputs.
 *
 * @param p     Port number (0 - 5)
 * @param v     Value (0 to 255 == (0mS - 512mS)
 * @return      good response or error message.
 *
 *****************************************************************************/

char * setPowerDelay(byte p, byte v) {
    if (p > 5) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Port %d\n", p);
        return rsp_buffer;
    }
    if (EPR_DATA.power_delays[p] != v) {
        EPR_DATA.power_delays[p] = v;
        epr_data_dirty = true;
    }
    return good_response;
}

/******************************************************************************
 *
 * getPowerDelay        Get the delay before powering up this pin.
 *
 * @param p     Port number (0 - 5)
 * @return      good response or error message.
 *
 *****************************************************************************/

char * getPowerDelay(byte p) {
    if (p > 5) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Port %d\n", p);
        return rsp_buffer;
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%03d\n", EPR_DATA.power_delays[p]);
    return rsp_buffer;
}

/******************************************************************************
 *
 * _get_current_htr     Get the current for the specified heater.
 *
 * @param h     Heater number (0-1)
 * @return      float amperage
 *
 ******************************************************************************/

float _get_current_htr(byte h) {
    float amps;
    byte old_pwm;

    if (EPR_DATA.pwm_values[h] == 0) {
        return 0.0;
    }
    old_pwm = EPR_DATA.pwm_values[h];
    EPR_DATA.pwm_values[h] = 255;

    delay(65);
    amps = analogRead(current_htr_pins[h]);
    delay(2);
    amps += analogRead(current_htr_pins[h]);
    delay(2);
    amps += analogRead(current_htr_pins[h]);
    delay(2);
    amps += analogRead(current_htr_pins[h]);
    EPR_DATA.pwm_values[h] = old_pwm;
    amps *= 0.25;
    amps = amps * 0.00292969 * calibration_mult_htr[h] + calibration_add_htr[h];
    if (amps < 0.030)
        amps = 0.0;
    amps *= (old_pwm / 255.0);
    current_htr[h] = amps;
    ah_htr[h] += amps / 360.0;
    return amps;
}

/******************************************************************************
 *
 * @param p switched port number
 * @return float amperage
 *
 *****************************************************************************/

float _get_current_pwr(byte p) {
    float amps;

    amps = analogRead(current_pwr_pins[p]);
    delay(2);
    amps += analogRead(current_pwr_pins[p]);
    delay(2);
    amps += analogRead(current_pwr_pins[p]);
    delay(2);
    amps += analogRead(current_pwr_pins[p]);
    amps *= 0.25;
    amps = amps * 0.00292969 * calibration_mult_pwr[p] + calibration_add_pwr[p];
    if (amps < 0.030)
        amps = 0.0;
    current_pwr[p] = amps;
    ah_pwr[p] += amps / 360.0;
    return amps;
}

/******************************************************************************
 *
 * _get_voltage()       Get the applied voltage (12V or so)
 *
 * @return float voltage
 *
 *****************************************************************************/

 float _get_voltage() {
    float volts;

    volts = analogRead(voltage_pin);
    delay(2);
    volts += analogRead(voltage_pin);
    delay(2);
    volts += analogRead(voltage_pin);
    delay(2);
    volts += analogRead(voltage_pin);
    voltage = volts / 4 * 0.004511899;  // 18V/4096, then calibrated.
    return voltage;
}

/******************************************************************************
 *
 * getPwrCurrent    Get the current used by a port.
 *
 * @param p     Port number (0 - 5)
 * @return      good response or error message.
 *
 * 3.3V = 4095
 * 3.3V = 33Amps
 * 1 ADU = 33/4096 Amps = 0.008056641 A per ADU
 * 3 samples: (ADU x 3) / 0.024169922 = Amps.
 *
 *****************************************************************************/

char * getPwrCurrent(byte p) {
    float amps;

    if (p > 5) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Port %d\n", p);
        return rsp_buffer;
    }
    amps = current_pwr[p];
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", amps);
    return rsp_buffer;
}

/******************************************************************************
 *
 * getHtrCurrent    Get the current used by a dew heater.
 *
 * @param h     Heater number (0 - 1)
 * @return      good response or error message.
 *
 *****************************************************************************/

char * getHtrCurrent(byte h) {
    float amps;
    byte old_pwm;

    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    amps = current_htr[h];
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", amps);
    return rsp_buffer;
}

/******************************************************************************
 *
 * getCurrentTotal  API call to get total current
 *
 * @return          float amperage
 *
 *****************************************************************************/

char * getCurrentTotal() {
    float amps = 0.0;

    for (int x = 0; x < 6; ++x) {
        amps += current_pwr[x];
    }
    for (int x = 0; x < 2; ++x) {
        amps += current_htr[x];
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", amps);
    return rsp_buffer;
}

char * getAHTotal() {
    float ah = 0.0;

    for (int x = 0; x < 6; ++x) {
        ah += ah_pwr[x];
    }
    for (int x = 0; x < 2; ++x) {
        ah += ah_htr[x];
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", ah);
    return rsp_buffer;
}

char * getHeaterAH(byte h) {
    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", ah_htr[h]);
    return rsp_buffer;
}

char * setHeaterAH(byte h, float value) {
    if (h > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", h);
        return rsp_buffer;
    }
    ah_htr[h] = value;
    return good_response;
}

char * getPortAH(byte p) {
    if (p > 5) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Port %d\n", p);
        return rsp_buffer;
    }
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", ah_pwr[p]);
    return rsp_buffer;
}

char * setPortAH(byte p, float value) {
    if (p > 1) {
        snprintf(rsp_buffer, sizeof(rsp_buffer), "ER:Htr %d\n", p);
        return rsp_buffer;
    }
    ah_pwr[p] = value;
    return good_response;
}

char * getRunTime() {

    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%03u:%02d:%02d\n",
             rt_hours, rt_minutes, rt_seconds);
    return rsp_buffer;
}

char * getADCValues(byte p) {
    float amps;
    int adu;
    adu = analogRead(current_pwr_pins[p]);
    amps = adu;
    delay(2);
    amps += analogRead(current_pwr_pins[p]);
    delay(2);
    amps += analogRead(current_pwr_pins[p]);
    amps *= 0.000976563;

    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%04d  %4.2f\n", adu, amps);
    return rsp_buffer;
}

/******************************************************************************
 *
 * getVoltageInput  API call to get input voltage (9V-14V)
 *
 * @return          float voltage
 *
 *****************************************************************************/

char * getVoltageInput() {
    float volts;
    
    volts = _get_voltage();
    snprintf(rsp_buffer, sizeof(rsp_buffer), "OK:%4.2f\n", volts);
    return rsp_buffer;
}

/******************************************************************************
 *
 * getPort      Gets the port number from the host message.
 *
 * @return      True on success, false otherwise.
 *
 *****************************************************************************/

bool getPort() {
    port = cmd_buffer[CMD_PORT] - '0';
    if (port < 0 || port > 5)
        return false;
    return true;
}

/******************************************************************************
 *
 * getPortAndValue  Get the channel number and value from the message.
 *
 * @param d         The type of value to expect. bool, byte, byte.
 * @return          True on success, false otherwise.
 *
 *****************************************************************************/

bool getPortAndValue(int d) {
    // Port is one digit, and value is a bool, byte or float.
    port = cmd_buffer[CMD_PORT] - '0';
    if (port < 0 || port > 5)
        return false;
    int tmp;
    switch (d) {
        case DT_BOOL:
            value_bool = cmd_buffer[CMD_VALUE] & 1;
            break;
        case DT_BYTE:
            tmp = atoi((const char *) &cmd_buffer[CMD_VALUE]);
            if (tmp > 255) {tmp = 255;}
            else if (tmp < 0) {tmp = 0;}
            value_byte = (byte)tmp;
            break;
        case DT_FLOAT:
            value_float = atof((const char *) &cmd_buffer[CMD_VALUE]);
            break;
        default:
            return false;
    }
    return true;
}

/******************************************************************************
 *
 * parseCommand     Takes the command apart and calls the appropriate action.
 *
 * @return          None.
 *
 *****************************************************************************/

void parseCommand() {
    if (!strncmp(cmd_buffer, "PWR", 3)) {
        // Command PWR sets or gets the state to ON (true) or OFF (false).
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BOOL)) {
                Serial.print(setPortPwr(port, value_bool));
            } else {
                Serial.print("ER:port or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPortPwr(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "PWA", 3)) {
        // Command PWA sets or gets the auto on at power up flag.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BOOL)) {
                Serial.print(setPortAuto(port, value_bool));
            } else {
                Serial.print("ER:port or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPortAuto(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "HTR", 3)) {
        // Command HTR sets or gets the PWM current value for the dew heater.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BYTE)) {
                Serial.print(setHeaterPwm(port, value_byte));
            } else {
                Serial.print("ER:htr or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getHeaterPwm(port));
            } else {
                Serial.print("ER:htr\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "HTA", 3)) {
        // Command HTA sets or gets the manual (false) or auto (true) flag for the heater.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BOOL)) {
                Serial.print(setHeaterAuto(port, value_bool));
            } else {
                Serial.print("ER:htr or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getHeaterAuto(port));
            } else {
                Serial.print("ER:htr\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "HTM", 3)) {
        // Command HTM sets or gets the heater minimum on value.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BYTE)) {
                Serial.print(setHeaterMinPwm(port, value_byte));
            } else {
                Serial.print("ER:htr or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getHeaterMinPwm(port));
            } else {
                Serial.print("ER:htr\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "HTG", 3)) {
        // Command HTA sets or gets the heater proportional gain.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BYTE)) {
                Serial.print(setHeaterGain(port, value_byte));
            } else {
                Serial.print("ER:htr or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getHeaterGain(port));
            } else {
                Serial.print("ER:htr\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "DLY", 3)) {
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BYTE)) {
                Serial.print(setPowerDelay(port, value_byte));
            } else {
                Serial.print("ER:port or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPowerDelay(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "HTO", 3)) {
        // Command HTO sets or gets the heater offset from dew point.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_BYTE)) {
                Serial.print(setHeaterOffset(port, value_byte));
            } else {
                Serial.print("ER:htr or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getHeaterOffset(port));
            } else {
                Serial.print("ER:htr\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "HAH", 3)) {
        // Command HAH sets or gets the heater amp hours accumulator.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_FLOAT)) {
                Serial.print(setHeaterAH(port, value_float));
            } else {
                Serial.print("ER:htr or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getHeaterAH(port));
            } else {
                Serial.print("ER:htr\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "PAH", 3)) {
        // Command HTO sets or gets the switched port amp hours accumulator.
        if (cmd_buffer[CMD_TYPE] == '#') {
            if (getPortAndValue(DT_FLOAT)) {
                Serial.print(setPortAH(port, value_float));
            } else {
                Serial.print("ER:htr or val\n");
            }
        } else if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPortAH(port));
            } else {
                Serial.print("ER:htr\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }

    } else if (!strncmp(cmd_buffer, "ADC", 3)) {
        // Get the current temperature.
        if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getADCValues(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "TMP", 3)) {
        // Get the current temperature.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(GetCurrentTemp());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "DPT", 3)) {
        // Get the current dew point.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(GetCurrentDewPoint());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "HUM", 3)) {
        // Get the current humidity.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(GetCurrentHumidity());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "VER", 3)) {
        // Get the program version number.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(version_string);
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "LNT", 3)) {
        // Get the current lens temperature.
        if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(GetLensTemp(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "CUR", 3)) {
        // Get the program version number.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(getCurrentTotal());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "AHT", 3)) {
        // Get the program version number.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(getAHTotal());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "VIN", 3)) {
        // Get the program version number.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(getVoltageInput());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "PWC", 3)) {
        // Get the current used by port temperature.
        if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getPwrCurrent(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "HTC", 3)) {
        // Get the current used by port temperature.
        if (cmd_buffer[CMD_TYPE] == '?') {
            if (getPort()) {
                Serial.print(getHtrCurrent(port));
            } else {
                Serial.print("ER:port\n");
            }
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "EPR", 3)) {
        // Dump the current EEPROM config.
        if (cmd_buffer[CMD_TYPE] == '?') {
            print_eeprom(&EPR_DATA);
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "RNT", 3)) {
        // Dump the current runtime.
        if (cmd_buffer[CMD_TYPE] == '?') {
            Serial.print(getRunTime());
        } else {
            Serial.print("ER:cmd type\n");
        }
    } else if (!strncmp(cmd_buffer, "DMP", 3)) {
        // Dump about everything.
        if (cmd_buffer[CMD_TYPE] == '?') {
            dump();
        } else if (cmd_buffer[CMD_TYPE] == '+') {
            dump_enabled = true;
        } else if (cmd_buffer[CMD_TYPE] == '-') {
            dump_enabled = false;
        } else {
            Serial.print("ER:cmd type\n");
        }

    } else {
        Serial.print("ER:unknown cmd\n");
    }
}

void dump() {
    memset(rsp_buffer, 0, sizeof(rsp_buffer));
    float tmpdiff1 = lens_temperature[0] - dew_point - EPR_DATA.heater_offset[0];
    float tmpdiff2 = lens_temperature[1] - dew_point - EPR_DATA.heater_offset[1];
    snprintf(rsp_buffer, 95, "%5.2f   %5.2f   %5.2f   %5.2f   %5.2f   %5.2f   %5.2f   %03u  %03u  %02x %02x %02x\n",
             temperature, humidity, dew_point, lens_temperature[0], lens_temperature[1], tmpdiff1,
             tmpdiff2, (int)EPR_DATA.pwm_values[0], (int)EPR_DATA.pwm_values[1], EPR_DATA.options, EPR_DATA.options2, EPR_DATA.options3);
    Serial.print(rsp_buffer);
}

/******************************************************************************
 *
 * calcDewPoint     Set the dew point based on the temperature and humidity.
 *
 * Td = T - ((100 – RH)/5)
 *
 *****************************************************************************/

void calcDewPoint() {
    dew_point = temperature - ((100 - humidity) / 5.0);
}

/******************************************************************************
 *
 * adjustHeaterPwm      Set the PWM based on temperature and dew point.
 *
 * @param h     Heater number (0 or 1)
 *
 *****************************************************************************/

void adjustHeater(byte h) {
    int pwm_value;
    if (!bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_1 + h)) {
        return;
    }
    if (!bitRead(EPR_DATA.options2, OPT_AUTO_DEW_1 + h)) {
        return;
    }
    float diff = lens_temperature[h] - dew_point - EPR_DATA.heater_offset[h];
    pwm_value = EPR_DATA.pwm_values[h];
    if (diff <= 0) {
        pwm_value = EPR_DATA.pwm_min_values[h] + (int) (abs(diff) * EPR_DATA.heater_gain[h]);
        if (pwm_value > 255) {
            pwm_value = 255;
        }
    }
    if (diff > 0) {
        pwm_value = 0;
    }
    EPR_DATA.pwm_values[h] = (byte) pwm_value;
}

/******************************************************************************
 *
 * reconcile_htr_auto   Figure out whether it should be on or off based on the
 * presence of the temperature probe.
 *
 *****************************************************************************/

bool reconcile_htr_auto(byte htr) {

    if (htr == 0) {
        if (!lens_temp1.begin(0x18)) {
            return false;
        } else {
            lens_temp1.setResolution(2);
            return true;
        }
    } else {
        if (!lens_temp2.begin(0x19)) {
            return false;
        } else {
            lens_temp2.setResolution(2);
            return true;
        }
    }
}

/******************************************************************************
 *
 *  pwm_process     Pseudo PWM with high side switches. Called from the
 *                  timer ISR.
 *
 *****************************************************************************/

void pwm_process() {
    noInterrupts();
    if (bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_1)) {
        if (pwm_loop == 0 && EPR_DATA.pwm_values[0] > 0) {
            digitalWrite(heater_pins[0], HIGH);
        }
        if (pwm_loop == EPR_DATA.pwm_values[0]) {
            digitalWrite(heater_pins[0], LOW);
        }
    }
    if (bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_2)) {
        if (pwm_loop == 0 && EPR_DATA.pwm_values[1] > 0) {
            digitalWrite(heater_pins[1], HIGH);
        }
        if (pwm_loop == EPR_DATA.pwm_values[1]) {
            digitalWrite(heater_pins[1], LOW);
        }
    }
    interrupts();
}

/******************************************************************************
 *
 * print_eeprom      Prints the EEPROM block to the Serial port.
 *
 *****************************************************************************/

void print_eeprom(Epr_Data * ptr) {
    byte x;
    Serial.print("SizeOf :  ");
    Serial.println(sizeof(EPR_DATA));
    Serial.print("SeqNo  :  ");
    Serial.println(ptr->seq_no);
    Serial.print("Sig    :  ");
    Serial.println(ptr->sig, 16);
    Serial.print("Opt    :  ");
    Serial.println(ptr->options, 16);
    Serial.print("Opt2   :  ");
    Serial.println(ptr->options2, 16);
    Serial.print("Opt3   :  ");
    Serial.println(ptr->options3, 17);
    Serial.print("Offset :  ");
    for (x=0;x<2;++x) {
        Serial.print(ptr->heater_offset[x]);
        Serial.print("  ");
    }
    Serial.println("");
    Serial.print("PWM min:  ");
    for (x=0;x<2;++x) {
        Serial.print(ptr->pwm_min_values[x]);
        Serial.print("  ");
    }
    Serial.println("");
    Serial.print("Gain   :  ");
    for (x=0;x<2;++x) {
        Serial.print(ptr->heater_gain[x]);
        Serial.print("  ");
    }
    Serial.println("");
    Serial.print("PWM    :  ");
    for (x=0;x<2;++x) {
        Serial.print(ptr->pwm_values[x]);
        Serial.print("  ");
    }
    Serial.println("");
    Serial.print("Delays :  ");
    for (x=0;x<6;++x) {
        Serial.print(ptr->power_delays[x]);
        Serial.print("  ");
    }
    Serial.println("");
}

/******************************************************************************
 *
 * setup        Setup the hardware and firmware to run.
 *
 *****************************************************************************/

void setup() {

    Serial.begin(230400);
    pinMode(LED, OUTPUT);

    Wire.begin();
    Wire.setClock(400000);
    Wire.setSDA(18);
    Wire.setSCL(19);

    TIMER.begin(timerRtn, TIMER_TIME);
    interrupts();
    load_eeprom();
    sensor.begin();

    // Check for dew temperature devices. Disable auto mode if not found.
    byte tmp = EPR_DATA.options2;
    EPR_DATA.options2 = bitSet(EPR_DATA.options2, OPT_DEW_ENABLE_1);
    if (reconcile_htr_auto(0)) {
        EPR_DATA.options2 = bitSet(EPR_DATA.options2, OPT_AUTO_DEW_1);
    } else {
        EPR_DATA.options2 = bitClear(EPR_DATA.options2, OPT_AUTO_DEW_1);
    }
    EPR_DATA.options2 = bitSet(EPR_DATA.options2, OPT_DEW_ENABLE_2);
    if (reconcile_htr_auto(1)) {
        EPR_DATA.options2 = bitSet(EPR_DATA.options2, OPT_AUTO_DEW_2);
    } else {
        EPR_DATA.options2 = bitClear(EPR_DATA.options2, OPT_AUTO_DEW_2);
    }
    if (EPR_DATA.options2 != tmp)
        epr_data_dirty = true;

    // Set the starting value of all ports, with delays.
    for (byte x = 0; x < 6; ++x) {
        pinMode(power_pins[x], OUTPUT);
        digitalWrite(power_pins[x], LOW);
        if (bitRead(EPR_DATA.options, x)) {
            delay(EPR_DATA.power_delays[x] * 4);
            digitalWrite(power_pins[x], HIGH);
            state |= bit(x);
        } else {
            state &= ~bit(x);
        }
    }
    for (byte x = 0; x < 2; ++x) {
        pinMode(heater_pins[x], OUTPUT);
        digitalWrite(heater_pins[x], LOW);
    }
    // Read the sensors to get initial values.
    humidity = sensor.readHumidity();
    temperature = sensor.readTemperature();
    if (humidity == 998.0) {
        temp_enabled = false;
        humidity = 0.0;
        temperature = 0.0;
    }
    if (bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_1)) {
        lens_temp1.wake();
        lens_temp1.setResolution(2);
    }
    if (bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_2)) {
        lens_temp2.wake();
        lens_temp2.setResolution(2);
    }
    analogReference(1);
    analogReadRes(12);
    memset(cmd_buffer, 0, sizeof(cmd_buffer));
    timer_tick = millis();
    eeprom_tick = millis() + EPR_TIME_DELAY;
    amperage_tick = millis() + AMP_TIME_DELAY;
    clock_tick = millis() + 1000;
    rt_hours = 0;
    rt_minutes = 0;
    rt_seconds = 0;
}

/******************************************************************************
 * loop     Main program loop.
 *
 * Checks for a command every 4mS, and gets the temperature and humidity
 * about once every second.
 *
 *****************************************************************************/

void loop() {

    // Did the host request something?
    if (Serial.available()) {
        // Read until linefeed.
        if (Serial.readBytesUntil(10, cmd_buffer, sizeof(cmd_buffer))) {
            parseCommand();
            memset(cmd_buffer, 0, sizeof(cmd_buffer));
        }
    }
    // Is it time to read humidity?
    ++loop_count;
    if (temp_enabled) {
        if (loop_count == 63) {
            humidity = sensor.readHumidity();
        }
        // Is it time to read temperature?
        if (loop_count == 125) {
            temperature = sensor.readTemperature();
        }
        // Is it time to lens temperature?
        if (loop_count == 25) {
            if (bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_1)) {
                lens_temperature[0] = lens_temp1.readTempC();
            }
            if (bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_2)) {
                lens_temperature[1] = lens_temp2.readTempC();
            }
        }
        // Is it time to calculate the heater set points?
        if (loop_count == 0) {
            calcDewPoint();
            if (EPR_DATA.options2 & OPT_AUTO_DEW_1) {
                adjustHeater(0);
            }
            if (EPR_DATA.options2 & OPT_AUTO_DEW_2) {
                adjustHeater(1);
            }
        }
    }
    if (amperage_tick < millis()) {
        amperage_tick = millis() + AMP_TIME_DELAY;
        for (byte idx = 0; idx < 6; ++idx) {
            current_pwr[idx] = _get_current_pwr(idx);
        }
        if (bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_1) && EPR_DATA.pwm_values[0] > 0) {
            current_htr[0] = _get_current_htr(0);
        } else {
            current_htr[0] = 0.0;
        }
        if (bitRead(EPR_DATA.options2, OPT_DEW_ENABLE_2) && EPR_DATA.pwm_values[1] > 0) {
            current_htr[1] = _get_current_htr(1);
        } else {
            current_htr[1] = 0.0;
        }
    }
    if (timer_tick < millis()) {
        timer_tick = millis() + DMP_TIME_DELAY;
        if (dump_enabled) {
            dump();
        }
    }
    if (eeprom_tick < millis()) {
        eeprom_tick = millis() + EPR_TIME_DELAY;
        if (epr_data_dirty) {
            store_eeprom();
            epr_data_dirty = false;
        }
    }
    if (clock_tick < millis()) {
        clock_tick = millis() + 1000;
        if(++rt_seconds > 59) {
            rt_seconds = 0;
            if (++rt_minutes > 59) {
                rt_minutes = 0;
                ++rt_hours;
            }
        }
    }
}

/******************************************************************************
 *
 * timerRtn     ISR for the timer. Does low speed PWM.
 *
 *****************************************************************************/

void timerRtn() {
    ++pwm_loop;
    pwm_process();
}
